
//-------------------------------------------------------------------------------//
// Author: Stefan Lörwald, Universität Heidelberg                                //
// License: CC BY-NC 4.0 http://creativecommons.org/licenses/by-nc/4.0/legalcode //
//-------------------------------------------------------------------------------//

#define COMPILE_TEMPLATE_ALGORITHM_CLASSES
#include "algorithm_classes.h"
#undef COMPILE_TEMPLATE_ALGORITHM_CLASSES

#include <algorithm>
#include <cassert>
#include <cstddef>
#include <set>
#include <tuple>
#include <vector>

#include "algorithm_map_operations.h"
#include "algorithm_row_operations.h"

template <typename Integer>
Facet<Integer> algorithm::facetClassRepresentative(const Facet<Integer>& facet, const Maps& maps)
{
   const auto facets = facetClass(facet, maps);
   assert( !facets.empty() );
   return *facets.crbegin();
}

template <typename Integer>
Vertex<Integer> algorithm::vertexClassRepresentative(const Vertex<Integer>& vertex, const Maps& maps)
{
   const auto vertices = vertexClass(vertex, maps);
   assert( !vertices.empty() );
   return *vertices.crbegin();
}

template <typename Integer>
std::set<Facet<Integer>> algorithm::facetClass(const Facet<Integer>& facet, const Maps& maps)
{
   std::set<Facet<Integer>> facets;
   facets.insert(facet);
   using Iterator = typename std::set<Facet<Integer>>::iterator;
   std::vector<Iterator> iterators;
   iterators.push_back(facets.begin());
   while ( !iterators.empty() )
   {
      const auto& current_facet = *iterators.back();
      iterators.pop_back();
      for ( const auto& map : maps )
      {
         const auto new_facet = applyOnInequality(map, current_facet);
         Iterator iterator;
         bool inserted;
         std::tie(iterator, inserted) = facets.insert(new_facet);
         if ( inserted )
         {
            iterators.push_back(iterator);
         }
      }
   }
   return facets;
}

template <typename Integer>
std::set<Vertex<Integer>> algorithm::vertexClass(const Vertex<Integer>& vertex, const Maps& maps)
{
   std::set<Vertex<Integer>> vertices;
   vertices.insert(vertex);
   using Iterator = typename std::set<Vertex<Integer>>::iterator;
   std::vector<Iterator> iterators;
   iterators.push_back(vertices.begin());
   while ( !iterators.empty() )
   {
      const auto& current_vertex = *iterators.back();
      iterators.pop_back();
      for ( const auto& map : maps )
      {
         const auto new_vertex = applyOnVertex(map, current_vertex);
         Iterator iterator;
         bool inserted;
         std::tie(iterator, inserted) = vertices.insert(new_vertex);
         if ( inserted )
         {
            iterators.push_back(iterator);
         }
      }
   }
   return vertices;
}

template <typename Integer>
Facets<Integer> algorithm::facetClasses(Facets<Integer> fs, const Maps& maps)
{
   std::set<Facet<Integer>> facets(fs.cbegin(), fs.cend());
   return facetClasses(facets, maps);
}

template <typename Integer>
Vertices<Integer> algorithm::vertexClasses(Vertices<Integer> vs, const Maps& maps)
{
   std::set<Vertex<Integer>> vertices(vs.cbegin(), vs.cend());
   return vertexClasses(vertices, maps);
}

template <typename Integer>
Facets<Integer> algorithm::facetClasses(std::set<Facet<Integer>> facets, const Maps& maps)
{
   Facets<Integer> classes;
   while ( !facets.empty() )
   {
      const auto facet_class = facetClass(*facets.begin(), maps);
      assert( !facet_class.empty() );
      classes.push_back(*facet_class.crbegin());
      for ( const auto& facet : facet_class )
      {
         const auto position = facets.find(facet);
         if ( position != facets.end() )
         {
            facets.erase(position);
         }
      }
   }
   return classes;
}

template <typename Integer>
Vertices<Integer> algorithm::vertexClasses(std::set<Vertex<Integer>> vertices, const Maps& maps)
{
   Vertices<Integer> classes;
   while ( !vertices.empty() )
   {
      const auto vertex_class = algorithm::vertexClass(*vertices.begin(), maps);
      assert( !vertex_class.empty() );
      classes.push_back(*vertex_class.crbegin());
      for ( const auto& vertex : vertex_class )
      {
         const auto position = vertices.find(vertex);
         if ( position != vertices.end() )
         {
            vertices.erase(position);
         }
      }
   }
   return classes;
}

