
//-------------------------------------------------------------------------------//
// Author: Stefan Lörwald, Universität Heidelberg                                //
// License: CC BY-NC 4.0 http://creativecommons.org/licenses/by-nc/4.0/legalcode //
//-------------------------------------------------------------------------------//

#include "input.h"

#include <algorithm>
#include <cassert>
#include <cstring>
#include <fstream>
#include <limits>
#include <stdexcept>
#include <string>

#include "algorithm_classes.h"
#include "algorithm_map_operations.h"
#include "algorithm_row_operations.h"
#include "input_common.h"
#include "input_consistency.h"
#include "input_detection.h"
#include "input_dimension.h"
#include "input_equation.h"
#include "input_inequality.h"
#include "input_keywords.h"
#include "input_map.h"
#include "input_names.h"
#include "input_order.h"
#include "input_validity.h"
#include "input_vertex.h"
#include "istream_peek_line.h"

namespace
{
   void sort(int, char**, Matrix<int>&);
   void expandInequalities(Matrix<int>&, const Maps&);
   void expandVertices(Matrix<int>&, const Maps&);

   std::string getFilenameKnownFacets(int, char**);
   std::string getFilenameKnownVertices(int, char**);
   Inequalities<int> knownFacets(int, char**, const Names&);
   Inequalities<int> knownVertices(int, char**);
}

template <>
std::tuple<Vertices<int>, Names, Maps, Inequalities<int>> input::vertices<int>(int argc, char** argv)
{
   /// In this method, we expect a file containing a conv/cone description in PANDA format.
   /// Raw format is not supported as deduction of which one is the convex hull part is
   /// impossible from data. We additionally may have names and maps.
   const auto filename = getFilename(argc, argv);
   std::ifstream file(filename.c_str());
   if ( !file )
   {
      throw std::invalid_argument("Failed to open file \"" + filename + "\".");
   }
   ConvexHull<int> conv;
   ConvexHull<int> cone;
   Names names;
   Maps maps;
   std::size_t dimension = std::numeric_limits<std::size_t>::max();
   for ( std::string token; advanceToNextKeyword(file, token); )
   {
      if ( implementation::isKeywordDimension(token) )
      {
         dimension = implementation::dimension(file);
      }
      else if ( implementation::isKeywordConvexHull(token) )
      {
         conv = implementation::verticesConvex(file);
         if ( checkValidity(argc, argv) )
         {
            implementation::checkValidityOfVertexClasses(conv, maps);
         }
      }
      else if ( implementation::isKeywordReducedConvexHull(token) )
      {
         conv = implementation::verticesConvex(file);
         if ( checkValidity(argc, argv) )
         {
            std::cerr << "Warning: validity of input vertices cannot be verified for reduced input.\n";
         }
         expandVertices(conv, maps);
      }
      else if ( implementation::isKeywordConicalHull(token) )
      {
         cone = implementation::verticesConical(file);
         if ( checkValidity(argc, argv) )
         {
            implementation::checkValidityOfVertexClasses(cone, maps);
         }
      }
      else if ( implementation::isKeywordReducedConicalHull(token) )
      {
         cone = implementation::verticesConical(file);
         if ( checkValidity(argc, argv) )
         {
            std::cerr << "Warning: validity of input rays cannot be verified for reduced input.\n";
         }
         expandVertices(cone, maps);
      }
      else if ( implementation::isKeywordNames(token) )
      {
         names = implementation::names(file);
      }
      else if ( implementation::isKeywordMaps(token) )
      {
         maps = implementation::maps(file, names);
      }
      else
      {
         std::getline(file, token);
      }
   }
   implementation::checkConsistency(conv, cone, names, maps, dimension);
   sort(argc, argv, cone);
   sort(argc, argv, conv);
   Matrix<int> vertices;
   vertices.reserve(cone.size() + conv.size());
   vertices.insert(vertices.end(), cone.cbegin(), cone.cend());
   vertices.insert(vertices.end(), conv.cbegin(), conv.cend());
   const auto known_facets = knownFacets(argc, argv, names);
   if ( input::checkValidity(argc, argv) )
   {
      input::implementation::checkValidityOfInequalities(vertices, known_facets);
   }
   return std::make_tuple(vertices, names, maps, known_facets);
}

template <>
std::tuple<Inequalities<int>, Names, Maps, Vertices<int>> input::inequalities<int>(int argc, char** argv)
{
   /// In this method we expect a file containing an outer description and possibly maps.
   const auto filename = getFilename(argc, argv);
   std::ifstream file(filename.c_str());
   if ( !file )
   {
      throw std::invalid_argument("Failed to open file \"" + filename + "\".");
   }
   Equations<int> equations;
   Inequalities<int> inequalities;
   std::size_t dimension = std::numeric_limits<std::size_t>::max();
   Maps maps;
   Names names;
   if ( !implementation::containsKeywords(file) )
   {
      throw std::invalid_argument("An outer description must be given in PANDA format.");
   }
   for ( std::string token; file && advanceToNextKeyword(file, token); )
   {
      if ( implementation::isKeywordDimension(token) )
      {
         dimension = implementation::dimension(file);
      }
      else if ( implementation::isKeywordMaps(token) )
      {
         maps = implementation::maps(file, names);
      }
      else if ( implementation::isKeywordNames(token) )
      {
         names = implementation::names(file);
      }
      else if ( implementation::isKeywordEquations(token) )
      {
         equations = implementation::equations(file, names);
      }
      else if ( implementation::isKeywordInequalities(token) )
      {
         inequalities = implementation::inequalitiesPanda(file, names);
         if ( checkValidity(argc, argv) )
         {
            implementation::checkValidityOfInequalityClasses(inequalities, maps);
         }
      }
      else if ( implementation::isKeywordReducedInequalities(token) )
      {
         inequalities = implementation::inequalitiesPanda(file, names);
         if ( checkValidity(argc, argv) )
         {
            std::cerr << "Warning: validity of input inequalities cannot be checked for reduced input.\n";
         }
         if ( equations.empty() )
         {
            std::cerr << "Warning: expanding the reduced input will be invalid, if the polytope is not full dimension (i.e. there are equations).\n";
         }
         const auto tmp_maps = algorithm::normalize(maps, equations);
         expandInequalities(inequalities, tmp_maps);
      }
      else
      {
         std::getline(file, token);
      }
   }
   implementation::checkConsistency(inequalities, names, dimension);
   inequalities.reserve(inequalities.size() + 2 * equations.size());
   for ( const auto& equation : equations )
   {
      inequalities.push_back(equation);
      inequalities.push_back(-1 * equation);
   }
   sort(argc, argv, inequalities);
   inequalities.emplace_back(inequalities.back().size(), 0);
   inequalities.back().back() = -1;
   const auto known_vertices = knownVertices(argc, argv);
   if ( input::checkValidity(argc, argv) )
   {
      input::implementation::checkValidityOfVertices(inequalities, known_vertices);
   }
   return std::make_tuple(inequalities, names, maps, known_vertices);
}

namespace
{
   std::string getFilenameKnownFacets(int argc, char** argv)
   {
      for ( int i(1); i < argc; ++i )
      {
         if ( std::strcmp(argv[i], "-k") == 0 )
         {
            if ( i + 1 == argc )
            {
               throw std::invalid_argument("Expected argument to option \"--known-facets\".");
            }
            return argv[i + 1];
         }
         if ( std::strncmp(argv[i], "--known-data=", 13) == 0 )
         {
            return argv[i] + 13;
         }
         if ( std::strncmp(argv[i], "--known-facets=", 15) == 0 )
         {
            return argv[i] + 15;
         }
         if ( std::strncmp(argv[i], "--known-vertices=", 17) == 0 )
         {
            throw std::invalid_argument("Unexpected argument \"--known-vertices\". Did you mean \"--known-data\"?");
         }
      }
      return "";
   }

   std::string getFilenameKnownVertices(int argc, char** argv)
   {
      for ( int i(1); i < argc; ++i )
      {
         if ( std::strcmp(argv[i], "-k") == 0 )
         {
            if ( i + 1 == argc )
            {
               throw std::invalid_argument("Expected argument to option \"--known-vertices\".");
            }
            return argv[i + 1];
         }
         if ( std::strncmp(argv[i], "--known-data=", 13) == 0 )
         {
            return argv[i] + 13;
         }
         if ( std::strncmp(argv[i], "--known-facets=", 15) == 0 )
         {
            throw std::invalid_argument("Unexpected argument \"--known-facets\". Did you mean \"--known-data\"?");
         }
         if ( std::strncmp(argv[i], "--known-vertices=", 17) == 0 )
         {
            return argv[i] + 17;
         }
      }
      return "";
   }

   Inequalities<int> knownFacets(int argc, char** argv, const Names& names)
   {
      const auto filename = getFilenameKnownFacets(argc, argv);
      if ( filename.empty() )
      {
         return {};
      }
      std::ifstream file(filename.c_str());
      if ( !file )
      {
         throw std::invalid_argument("Failed to open file \"" + filename + "\".");
      }
      std::string token;
      input::advanceToNextKeyword(file, token);
      if ( !input::implementation::isKeywordInequalities(token) )
      {
         throw std::invalid_argument("Expected no other keyword than \"Inequalities\" in a file containing known facets.");
      }
      return input::implementation::inequalitiesPanda(file, names);
   }

   Vertices<int> knownVertices(int argc, char** argv)
   {
      const auto filename = getFilenameKnownVertices(argc, argv);
      if ( filename.empty() )
      {
         return {};
      }
      std::ifstream file(filename.c_str());
      if ( !file )
      {
         throw std::invalid_argument("Failed to open file \"" + filename + "\".");
      }
      Vertices<int> conv;
      Vertices<int> cone;
      for ( std::string token; input::advanceToNextKeyword(file, token); )
      {
         if ( input::implementation::isKeywordConvexHull(token) ||
              input::implementation::isKeywordReducedConvexHull(token) )
         {
            conv = input::implementation::verticesConvex(file);
         }
         else if ( input::implementation::isKeywordConicalHull(token) ||
                   input::implementation::isKeywordReducedConicalHull(token) )
         {
            cone = input::implementation::verticesConical(file);
         }
         else
         {
            std::getline(file, token);
         }
      }
      conv.reserve(conv.size() + cone.size());
      conv.insert(conv.end(), cone.cbegin(), cone.cend());
      return conv;
   }
}

namespace
{
   void sort(int argc, char** argv, Matrix<int>& matrix)
   {
      const auto order = getInputOrder(argc, argv);
      switch ( order )
      {
         case InputOrder::LexicographicAscending:
         {
            std::sort(matrix.begin(), matrix.end());
            return;
         }
         case InputOrder::LexicographicDescending:
         {
            std::sort(matrix.rbegin(), matrix.rend());
            return;
         }
         case InputOrder::NonZeroEntriesAscending:
         {
            std::sort(matrix.begin(), matrix.end(), [](const Row<int>& a, const Row<int>& b)
            {
               const auto zero_entries_a = std::count(a.cbegin(), a.cend(), 0);
               const auto zero_entries_b = std::count(b.cbegin(), b.cend(), 0);
               if ( zero_entries_a != zero_entries_b )
               {
                  return (zero_entries_a > zero_entries_b);
               }
               return (a < b);
            });
            return;
         }
         case InputOrder::NonZeroEntriesDescending:
         {
            std::sort(matrix.begin(), matrix.end(), [](const Row<int>& a, const Row<int>& b)
            {
               const auto zero_entries_a = std::count(a.cbegin(), a.cend(), 0);
               const auto zero_entries_b = std::count(b.cbegin(), b.cend(), 0);
               if ( zero_entries_a != zero_entries_b )
               {
                  return (zero_entries_a < zero_entries_b);
               }
               return (a < b);
            });
            return;
         }
         case InputOrder::Reverse:
         {
            std::reverse(matrix.begin(), matrix.end());
            return;
         }
         case InputOrder::NoSorting:
         {
            return;
         }
         default:
         {
            throw std::invalid_argument("Invalid sorting command. (This is most likely a bug!)");
         }
      }
   }

   void expandInequalities(Matrix<int>& matrix, const Maps& maps)
   {
      if ( maps.empty() )
      {
         throw std::invalid_argument("The system of inequalities cannot be a reduced system without any maps. Maps must be declared before the section of inequalities.");
      }
      Matrix<int> all;
      for ( const auto& row : matrix )
      {
         const auto row_class = algorithm::facetClass(row, maps);
         all.reserve(all.size() + row_class.size());
         all.insert(all.end(), row_class.cbegin(), row_class.cend());
      }
      matrix = std::move(all);
   }

   void expandVertices(Matrix<int>& matrix, const Maps& maps)
   {
      if ( maps.empty() )
      {
         throw std::invalid_argument("The system of vertices / rays cannot be a reduced system without any maps. Maps must be declared before the section of vertices / rays.");
      }
      Matrix<int> all;
      for ( const auto& row : matrix )
      {
         const auto row_class = algorithm::vertexClass(row, maps);
         all.reserve(all.size() + row_class.size());
         all.insert(all.end(), row_class.cbegin(), row_class.cend());
      }
      matrix = std::move(all);
   }
}

