
#define COMPILE_TEMPLATE_LIST
#include "list.h"
#undef COMPILE_TEMPLATE_LIST

#include <cstddef>
#include <iostream>
#include <sstream>

#include "algorithm_row_operations.h"

#define PRINT_DONE_COUNTER /// if enabled, the beginning of processing a facet will be announced.

#ifdef PRINT_DONE_COUNTER
   #if __GNUC__ < 4  || (__GNUC__ == 4 && __GNUC_MINOR__ <= 7)
      // thread_local is not supported prior to gcc 4.8
      #include <map>
      #include <thread>
      namespace
      {
         std::map<std::thread::id, std::size_t> indices;
      }
      #define USE_INDEX_WORKAROUND
   #else
      namespace
      {
         thread_local std::size_t index;
      }
   #endif
#endif

template <typename Integer>
void List<Integer>::put(const Matrix<Integer>& rows, FacetTag) const
{
   for ( const auto& row : rows )
   {
      put(row, FacetTag{});
   }
   std::lock_guard<std::mutex> lock(mutex);
   --workers;
   #ifdef PRINT_DONE_COUNTER
   #ifdef USE_INDEX_WORKAROUND
   auto index = indices[std::this_thread::get_id()];
   #endif
   if ( index > 0 )
   {
      std::stringstream stream;
      stream << "Done processing #" << index << "\n";
      std::cerr << stream.str();
   }
   #endif
}

template <typename Integer>
void List<Integer>::put(const Matrix<Integer>& rows, VertexTag) const
{
   for ( const auto& row : rows )
   {
      put(row, VertexTag{});
   }
   std::lock_guard<std::mutex> lock(mutex);
   --workers;
   #ifdef PRINT_DONE_COUNTER
   #ifdef USE_INDEX_WORKAROUND
   auto index = indices[std::this_thread::get_id()];
   #endif
   if ( index > 0 )
   {
      std::stringstream stream;
      stream << "Done processing #" << index << "\n";
      std::cerr << stream.str();
   }
   #endif
}

template <typename Integer>
void List<Integer>::put(const Row<Integer>& facet, FacetTag) const
{
   std::lock_guard<std::mutex> lock(mutex);
   Iterator it;
   bool added;
   std::tie(it, added) = facets.insert(facet);
   if ( added )
   {
      prettyPrintln(std::cout, facet, names, "<=");
      std::cout.flush();
      iterators.push_back(it);
      condition.notify_one();
   }
}

template <typename Integer>
void List<Integer>::put(const Row<Integer>& facet, VertexTag) const
{
   std::lock_guard<std::mutex> lock(mutex);
   Iterator it;
   bool added;
   std::tie(it, added) = facets.insert(facet);
   if ( added )
   {
      std::cout << facet << '\n';
      std::cout.flush();
      iterators.push_back(it);
      condition.notify_one();
   }
}

template <typename Integer>
Row<Integer> List<Integer>::get() const
{
   if ( empty() ) // abort
   {
      const auto it = facets.insert(Row<Integer>{}).first;
      std::unique_lock<std::mutex> lock(mutex);
      iterators.push_back(it);
      condition.notify_all();
   }
   std::unique_lock<std::mutex> lock(mutex);
   condition.wait(lock, [&](){ return !iterators.empty(); });
   ++workers;
   const auto facet = *iterators.front();
   if ( !facet.empty() )
   {
      iterators.erase(iterators.begin());
   }
   #ifdef PRINT_DONE_COUNTER
   if ( !facet.empty() )
   {
      ++counter;
      #ifdef USE_INDEX_WORKAROUND
      indices[std::this_thread::get_id()] = counter;
      #else
      index = counter;
      #endif
      std::stringstream stream;
      stream << "Processing #" << counter << " of at least " << facets.size() << " classes\n";
      std::cerr << stream.str();
   }
   #endif
   return facet;
}

template <typename Integer>
List<Integer>::List(const Names& names_)
:
   names(names_),
   mutex(),
   workers(1),
   condition(),
   facets(),
   iterators(),
   counter(0)
{
}

template <typename Integer>
bool List<Integer>::empty() const
{
   const std::lock_guard<std::mutex> lock(mutex);
   return workers == 0 && iterators.empty();
}

