
//-------------------------------------------------------------------------------//
// Author: Stefan Lörwald, Universität Heidelberg                                //
// License: CC BY-NC 4.0 http://creativecommons.org/licenses/by-nc/4.0/legalcode //
//-------------------------------------------------------------------------------//

#define COMPILE_TEMPLATE_METHOD_ADJACENCY_DECOMPOSITION_IMPLEMENTATION
#include "method_adjacency_decomposition_implementation.h"
#undef COMPILE_TEMPLATE_METHOD_ADJACENCY_DECOMPOSITION_IMPLEMENTATION

#include <algorithm>
#include <future>
#include <iostream>
#include <list>

#include "algorithm_classes.h"
#include "algorithm_fourier_motzkin_elimination.h"
#include "algorithm_map_operations.h"
#include "algorithm_matrix_operations.h"
#include "algorithm_rotation.h"
#include "algorithm_row_operations.h"
#include "concurrency.h"
#include "joining_thread.h"
#include "message_passing_interface_session.h"

namespace
{
   template <typename Integer>
   std::pair<Equations<Integer>, Maps> reduce(const JobManager<Integer>&, const std::tuple<Matrix<Integer>, Names, Maps, Matrix<Integer>>& data, FacetTag);

   template <typename Integer>
   std::pair<Equations<Integer>, Maps> reduce(const JobManager<Integer>&, const std::tuple<Matrix<Integer>, Names, Maps, Matrix<Integer>>& data, VertexTag);

   template <typename Integer>
   std::pair<Equations<Integer>, Maps> reduce(const JobManagerProxy<Integer>&, const std::tuple<Matrix<Integer>, Names, Maps, Matrix<Integer>>& data, FacetTag);

   template <typename Integer>
   std::pair<Equations<Integer>, Maps> reduce(const JobManagerProxy<Integer>&, const std::tuple<Matrix<Integer>, Names, Maps, Matrix<Integer>>& data, VertexTag);

   template <typename Integer>
   std::future<void> initializePool(JobManager<Integer>&, const Matrix<Integer>&, const Maps&, const Matrix<Integer>&, const Equations<Integer>&, FacetTag);

   template <typename Integer>
   std::future<void> initializePool(JobManager<Integer>&, const Matrix<Integer>&, const Maps&, const Matrix<Integer>&, const Equations<Integer>&, VertexTag);

   template <typename Integer>
   std::future<void> initializePool(JobManagerProxy<Integer>&, const Matrix<Integer>&, const Maps&, const Matrix<Integer>&, const Equations<Integer>&, FacetTag);

   template <typename Integer>
   std::future<void> initializePool(JobManagerProxy<Integer>&, const Matrix<Integer>&, const Maps&, const Matrix<Integer>&, const Equations<Integer>&, VertexTag);
}

template <template <typename> class JobManagerType, typename Integer, typename TagType>
void implementation::adjacencyDecomposition(int argc, char** argv, const std::tuple<Matrix<Integer>, Names, Maps, Matrix<Integer>>& data, TagType tag)
{
   const auto node_count = getMessagePassingInterfaceSession().getNumberOfNodes();
   const auto thread_count = numberOfThreads(argc, argv);
   const auto& input = std::get<0>(data);
   const auto& names = std::get<1>(data);
   const auto& known_output = std::get<3>(data);
   JobManagerType<Integer> job_manager(names, node_count, thread_count, tag);
   const auto reduced_data = reduce(job_manager, data, tag);
   const auto& equations = std::get<0>(reduced_data);
   const auto& maps = std::get<1>(reduced_data);
   std::list<JoiningThread> threads;
   auto future = initializePool(job_manager, input, maps, known_output, equations, tag);
   for ( int i(0); i < thread_count; ++i )
   {
      threads.emplace_front([&]()
      {
         while ( true )
         {
            const auto job = job_manager.get();
            if ( job.empty() )
            {
               break;
            }
            const auto jobs = algorithm::rotation(input, job, maps, tag);
            job_manager.put(jobs, tag);
         }
      });
   }
   future.wait();
}

namespace
{
   template <typename Integer>
   std::pair<Equations<Integer>, Maps> reduce(const JobManager<Integer>&, const std::tuple<Matrix<Integer>, Names, Maps, Matrix<Integer>>& data, FacetTag)
   {
      const auto& vertices = std::get<0>(data);
      const auto& names = std::get<1>(data);
      const auto& original_maps = std::get<2>(data);
      const auto equations = algorithm::extractEquations(vertices);
      if ( !equations.empty() )
      {
         std::cout << "Equations:\n";
         prettyPrint(std::cout, equations, names, "=");
         std::cout << "\n";
         const auto maps = algorithm::normalize(original_maps, equations);
         return std::make_pair(equations, maps);
      }
      return std::make_pair(equations, original_maps);
   }

   template <typename Integer>
   std::pair<Equations<Integer>, Maps> reduce(const JobManager<Integer>&, const std::tuple<Matrix<Integer>, Names, Maps, Matrix<Integer>>& data, VertexTag)
   {
      const auto& original_maps = std::get<2>(data);
      return std::make_pair(Equations<Integer>{}, original_maps);
   }

   template <typename Integer>
   std::pair<Equations<Integer>, Maps> reduce(const JobManagerProxy<Integer>&, const std::tuple<Matrix<Integer>, Names, Maps, Matrix<Integer>>& data, FacetTag)
   {
      const auto& vertices = std::get<0>(data);
      const auto& original_maps = std::get<2>(data);
      const auto equations = algorithm::extractEquations(vertices);
      if ( !equations.empty() )
      {
         const auto maps = algorithm::normalize(original_maps, equations);
         return std::make_pair(equations, maps);
      }
      return std::make_pair(equations, original_maps);
   }

   template <typename Integer>
   std::pair<Equations<Integer>, Maps> reduce(const JobManagerProxy<Integer>&, const std::tuple<Matrix<Integer>, Names, Maps, Matrix<Integer>>& data, VertexTag)
   {
      const auto& original_maps = std::get<2>(data);
      return std::make_pair(Equations<Integer>{}, original_maps);
   }

   template <typename Integer>
   std::future<void> initializePool(JobManager<Integer>& manager, const Matrix<Integer>& matrix, const Maps& maps, const Matrix<Integer>& known_output, const Equations<Integer>& equations, FacetTag tag)
   {
      if ( !maps.empty() )
      {
         std::cout << "Reduced ";
      }
      std::cout << "Inequalities:\n";
      // Initialize the process (so that other processes start).
      if ( !known_output.empty() )
      {
         auto tmp = normalize(known_output.front(), equations);
         tmp = algorithm::facetClassRepresentative(tmp, maps);
         manager.put(Matrix<Integer>{tmp}, tag);
      }
      else
      {
         auto facets = algorithm::fourierMotzkinEliminationHeuristic(matrix);
         for ( auto& facet : facets )
         {
            facet = algorithm::facetClassRepresentative(facet, maps);
         }
         manager.put(facets, tag);
      }
      // Add the remaining known facets from file asynchronously.
      auto future = std::async(std::launch::async, [&]()
      {
         for ( const auto& facet : known_output )
         {
            auto tmp = normalize(facet, equations);
            tmp = algorithm::facetClassRepresentative(tmp, maps);
            manager.put(tmp, tag);
         }
      });
      return future;
   }

   template <typename Integer>
   std::future<void> initializePool(JobManager<Integer>& manager, const Matrix<Integer>& inequalities, const Maps& maps, const Matrix<Integer>& known_output, const Equations<Integer>&, VertexTag tag)
   {
      if ( !maps.empty() )
      {
         std::cout << "Reduced ";
      }
      std::cout << "Vertices / Rays:\n";
      // Initialize the process (so that other processes start).
      if ( !known_output.empty() )
      {
         const auto tmp = algorithm::vertexClassRepresentative(known_output.front(), maps);
         manager.put(Matrix<Integer>{tmp}, tag);
      }
      else
      {
         auto vertices = algorithm::fourierMotzkinEliminationHeuristic(inequalities);
         for ( auto& vertex : vertices )
         {
            vertex = algorithm::vertexClassRepresentative(vertex, maps);
         }
         manager.put(vertices, tag);
      }
      // Add the remaining known vertices / rays from file asynchronously.
      auto future = std::async(std::launch::async, [&]()
      {
         for ( const auto& vertex : known_output )
         {
            const auto tmp = algorithm::vertexClassRepresentative(vertex, maps);
            manager.put(tmp, tag);
         }
      });
      return future;
   }

   template <typename Integer>
   std::future<void> initializePool(JobManagerProxy<Integer>&, const ConvexHull<Integer>&, const Maps&, const Inequalities<Integer>&, const Equations<Integer>&, FacetTag)
   {
      // only the manager on the root node performs a heuristic to get initial facets.
      auto future = std::async(std::launch::async, [](){});
      return future;
   }

   template <typename Integer>
   std::future<void> initializePool(JobManagerProxy<Integer>&, const ConvexHull<Integer>&, const Maps&, const Inequalities<Integer>&, const Equations<Integer>&, VertexTag)
   {
      // only the manager on the root node performs a heuristic to get initial facets.
      auto future = std::async(std::launch::async, [](){});
      return future;
   }
}

