
//-------------------------------------------------------------------------------//
// Author: Stefan Lörwald, Universität Heidelberg                                //
// License: CC BY-NC 4.0 http://creativecommons.org/licenses/by-nc/4.0/legalcode //
//-------------------------------------------------------------------------------//

#include "testing_gear.h"

#include "input_inequality_unnamed_variables.h"

#include <sstream>
#include <stdexcept>

using input::implementation::inequality;

namespace
{
   void valid();
   void invalid();
}

int main()
try
{
   valid();
   invalid();
}
catch ( const std::exception& e )
{
   std::cerr << e.what() << "\n";
   return 1;
}

namespace
{
   void valid()
   {
      for ( const auto string : {"   1 \t2 -3  -10\t 1",
                                 "   1 \t2 -3  -10\t <= -1",
                                 "   1 \t2 -3  -10\t <= -1",
                                 "   1 \t2 -3  -10\t = -1",
                                 "   1 \t2 -3  -10\t == -1",
                                 "   -1 \t-2 3  10\t >= 1",
                                 "   -1 \t-2 3  10\t >= 1",
                                 " ( 1 )  1 \t2 -3  -10\t 1",
                                 " ( a b)  1 \t2 -3  -10\t <= -1",
                                 " ()  1 \t2 -3  -10\t <= -1",
                                 " (3 )  1 \t2 -3  -10\t = -1",
                                 " ( 44)  1 \t2 -3  -10\t == -1",
                                 " (5)  -1 \t-2 3  10\t >= 1",
                                 " (6)  -1 \t-2 3  10\t >= 1"} )
      {
         {
            std::istringstream stream(string);
            ASSERT_NOTHROW(inequality(stream), "On valid input, inequality read method may not throw.");
         }
         {
            std::istringstream stream(string);
            ASSERT((inequality(stream).first == Inequality<int>{1, 2, -3, -10, 1}), "Data mismatch.");
         }
      }
      {
         std::stringstream stream("");
         inequality(stream);
         ASSERT_NOTHROW(inequality(stream), "Empty line may not throw exception.");
      }
      {
         std::stringstream stream("");
         inequality(stream);
         ASSERT(inequality(stream).first.empty(), "Empty line must result in empty row.");
      }
   }

   void invalid()
   {
      for ( const auto string : {" 1 \t2 -3 -10 \t <=",
                                 " 1 \t2 -3 -10 \t <= ",
                                 " 1 \t2 -3 -10 \t < ",
                                 " 1 \t2 -3 -10 \t <",
                                 " 1 \t2 -3 -10 \t =",
                                 " 1 \t2 -3 -10 \t = ",
                                 " 1 \t2 -3 -10 \t ==",
                                 " 1 \t2 -3 -10 \t == ",
                                 " 1 \t2 -3 -10 \t >",
                                 " 1 \t2 -3 -10 \t > ",
                                 " 1 \t2 -3 -10 \t >=",
                                 " 1 \t2 -3 -10 \t >= ",
                                 " 1 \t2 -3 -10 \t >= \n 2"} )
      {
         std::stringstream stream(string);
         const auto message = "Invalid input must result in std::invalid_argument (missing right hand side).";
         ASSERT_EXCEPTION(inequality(stream), std::invalid_argument, message);
      }
      for ( const auto string : {"   1 \t2 -3  -10\t < 0",
                                 "   1 \t2 -3  -10\t < -1",
                                 "   1 \t2 -3  -10\t ? -1",
                                 "   -1 \t-2 3  10\t > 1",
                                 "   -1 \t-2 3  10\t > 0",
                                 " ( a b)  1 \t2 -3  -10\t < 0",
                                 " ()  1 \t2 -3  -10\t < -1",
                                 " (5)  -1 \t-2 3  10\t > 1",
                                 " (6)  -1 \t-2 3  10\t > 0"} )
      {
         {
            std::istringstream stream(string);
            ASSERT_ANY_EXCEPTION(inequality(stream), "If an invalid relation operator is used, an exception must be thrown.");
         }
      }
      for ( const auto string : {" 1 \t a 2",
                                 " (",
                                 "( 1 2 3"} )
      {
         std::stringstream stream(string);
         std::string message = "Invalid input must result in std::invalid_argument";
         message += " (invalid character or identifier).";
         ASSERT_EXCEPTION(inequality(stream), std::invalid_argument, message);
      }
   }
}

